from __future__ import absolute_import
import itertools

import lit.util
from lit.ShCommands import Command, GlobItem, Pipeline, Seq


class ShLexer:
    def __init__(self, data, win32Escapes=False):
        self.data = data
        self.pos = 0
        self.end = len(data)
        self.win32Escapes = win32Escapes

    def eat(self):
        c = self.data[self.pos]
        self.pos += 1
        return c

    def look(self):
        return self.data[self.pos]

    def maybe_eat(self, c):
        """
        maybe_eat(c) - Consume the character c if it is the next character,
        returning True if a character was consumed."""
        if self.data[self.pos] == c:
            self.pos += 1
            return True
        return False

    def lex_arg_fast(self, c):
        # Get the leading whitespace free section.
        chunk = self.data[self.pos - 1 :].split(None, 1)[0]

        # If it has special characters, the fast path failed.
        if (
            "|" in chunk
            or "&" in chunk
            or "<" in chunk
            or ">" in chunk
            or "'" in chunk
            or '"' in chunk
            or ";" in chunk
            or "\\" in chunk
        ):
            return None

        self.pos = self.pos - 1 + len(chunk)
        return GlobItem(chunk) if "*" in chunk or "?" in chunk else chunk

    def lex_arg_slow(self, c):
        if c in "'\"":
            str = self.lex_arg_quoted(c)
        else:
            str = c
        unquoted_glob_char = False
        quoted_glob_char = False
        while self.pos != self.end:
            c = self.look()
            if c.isspace() or c in "|&;":
                break
            elif c in "><":
                # This is an annoying case; we treat '2>' as a single token so
                # we don't have to track whitespace tokens.

                # If the parse string isn't an integer, do the usual thing.
                if not str.isdigit():
                    break

                # Otherwise, lex the operator and convert to a redirection
                # token.
                num = int(str)
                tok = self.lex_one_token()
                assert isinstance(tok, tuple) and len(tok) == 1
                return (tok[0], num)
            elif c == '"' or c == "'":
                self.eat()
                quoted_arg = self.lex_arg_quoted(c)
                if "*" in quoted_arg or "?" in quoted_arg:
                    quoted_glob_char = True
                str += quoted_arg
            elif not self.win32Escapes and c == "\\":
                # Outside of a string, '\\' escapes everything.
                self.eat()
                if self.pos == self.end:
                    lit.util.warning(
                        "escape at end of quoted argument in: %r" % self.data
                    )
                    return str
                str += self.eat()
            elif c in "*?":
                unquoted_glob_char = True
                str += self.eat()
            else:
                str += self.eat()
        # If a quote character is present, lex_arg_quoted will remove the quotes
        # and append the argument directly.  This causes a problem when the
        # quoted portion contains a glob character, as the character will no
        # longer be treated literally.  If glob characters occur *only* inside
        # of quotes, then we can handle this by not globbing at all, and if
        # glob characters occur *only* outside of quotes, we can still glob just
        # fine.  But if a glob character occurs both inside and outside of
        # quotes this presents a problem.  In practice this is such an obscure
        # edge case that it doesn't seem worth the added complexity to support.
        # By adding an assertion, it means some bot somewhere will catch this
        # and flag the user of a non-portable test (which could almost certainly
        # be re-written to work correctly without triggering this).
        assert not (quoted_glob_char and unquoted_glob_char)
        return GlobItem(str) if unquoted_glob_char else str

    def lex_arg_quoted(self, delim):
        str = ""
        while self.pos != self.end:
            c = self.eat()
            if c == delim:
                return str
            elif c == "\\" and delim == '"':
                # Inside a '"' quoted string, '\\' only escapes the quote
                # character and backslash, otherwise it is preserved.
                if self.pos == self.end:
                    lit.util.warning(
                        "escape at end of quoted argument in: %r" % self.data
                    )
                    return str
                c = self.eat()
                if c == '"':  #
                    str += '"'
                elif c == "\\":
                    str += "\\"
                else:
                    str += "\\" + c
            else:
                str += c
        lit.util.warning("missing quote character in %r" % self.data)
        return str

    def lex_arg_checked(self, c):
        pos = self.pos
        res = self.lex_arg_fast(c)
        end = self.pos

        self.pos = pos
        reference = self.lex_arg_slow(c)
        if res is not None:
            if res != reference:
                raise ValueError("Fast path failure: %r != %r" % (res, reference))
            if self.pos != end:
                raise ValueError("Fast path failure: %r != %r" % (self.pos, end))
        return reference

    def lex_arg(self, c):
        return self.lex_arg_fast(c) or self.lex_arg_slow(c)

    def lex_one_token(self):
        """
        lex_one_token - Lex a single 'sh' token."""

        c = self.eat()
        if c == ";":
            return (c,)
        if c == "|":
            if self.maybe_eat("|"):
                return ("||",)
            return (c,)
        if c == "&":
            if self.maybe_eat("&"):
                return ("&&",)
            if self.maybe_eat(">"):
                return ("&>",)
            return (c,)
        if c == ">":
            if self.maybe_eat("&"):
                return (">&",)
            if self.maybe_eat(">"):
                return (">>",)
            return (c,)
        if c == "<":
            if self.maybe_eat("&"):
                return ("<&",)
            if self.maybe_eat(">"):
                return ("<<",)
            return (c,)

        return self.lex_arg(c)

    def lex(self):
        while self.pos != self.end:
            if self.look().isspace():
                self.eat()
            else:
                yield self.lex_one_token()


###


class ShParser:
    def __init__(self, data, win32Escapes=False, pipefail=False):
        self.data = data
        self.pipefail = pipefail
        self.tokens = ShLexer(data, win32Escapes=win32Escapes).lex()

    def lex(self):
        for item in self.tokens:
            return item
        return None

    def look(self):
        token = self.lex()
        if token is not None:
            self.tokens = itertools.chain([token], self.tokens)
        return token

    def parse_command(self):
        tok = self.lex()
        if not tok:
            raise ValueError("empty command!")
        if isinstance(tok, tuple):
            raise ValueError("syntax error near unexpected token %r" % tok[0])

        args = [tok]
        redirects = []
        while 1:
            tok = self.look()

            # EOF?
            if tok is None:
                break

            # If this is an argument, just add it to the current command.
            if isinstance(tok, (str, GlobItem)):
                args.append(self.lex())
                continue

            # Otherwise see if it is a terminator.
            assert isinstance(tok, tuple)
            if tok[0] in ("|", ";", "&", "||", "&&"):
                break

            # Otherwise it must be a redirection.
            op = self.lex()
            arg = self.lex()
            if not arg:
                raise ValueError("syntax error near token %r" % op[0])
            redirects.append((op, arg))

        return Command(args, redirects)

    def parse_pipeline(self):
        negate = False

        commands = [self.parse_command()]
        while self.look() == ("|",):
            self.lex()
            commands.append(self.parse_command())
        return Pipeline(commands, negate, self.pipefail)

    def parse(self):
        lhs = self.parse_pipeline()

        while self.look():
            operator = self.lex()
            assert isinstance(operator, tuple) and len(operator) == 1

            if not self.look():
                raise ValueError("missing argument to operator %r" % operator[0])

            # FIXME: Operator precedence!!
            lhs = Seq(lhs, operator[0], self.parse_pipeline())

        return lhs



"""
lit - LLVM Integrated Tester.

See lit.pod for more information.
"""

import itertools
import os
import platform
import sys
import time

import lit.cl_arguments
import lit.discovery
import lit.display
import lit.LitConfig
import lit.reports
import lit.run
import lit.Test
import lit.util
from lit.formats.googletest import GoogleTest
from lit.TestTimes import record_test_times


def main(builtin_params={}):
    opts = lit.cl_arguments.parse_args()
    params = create_params(builtin_params, opts.user_params)
    is_windows = platform.system() == "Windows"

    lit_config = lit.LitConfig.LitConfig(
        progname=os.path.basename(sys.argv[0]),
        path=opts.path,
        quiet=opts.quiet,
        useValgrind=opts.useValgrind,
        valgrindLeakCheck=opts.valgrindLeakCheck,
        valgrindArgs=opts.valgrindArgs,
        noExecute=opts.noExecute,
        debug=opts.debug,
        isWindows=is_windows,
        order=opts.order,
        params=params,
        config_prefix=opts.configPrefix,
        echo_all_commands=opts.echoAllCommands,
    )

    discovered_tests = lit.discovery.find_tests_for_inputs(
        lit_config, opts.test_paths
    )
    if not discovered_tests:
        sys.stderr.write("error: did not discover any tests for provided path(s)\n")
        sys.exit(2)

    if opts.show_suites or opts.show_tests:
        print_discovered(discovered_tests, opts.show_suites, opts.show_tests)
        sys.exit(0)

    if opts.show_used_features:
        features = set(
            itertools.chain.from_iterable(
                t.getUsedFeatures()
                for t in discovered_tests
                if t.gtest_json_file is None
            )
        )
        print(" ".join(sorted(features)))
        sys.exit(0)

    # Command line overrides configuration for maxIndividualTestTime.
    if opts.maxIndividualTestTime is not None:  # `not None` is important (default: 0)
        if opts.maxIndividualTestTime != lit_config.maxIndividualTestTime:
            lit_config.note(
                (
                    "The test suite configuration requested an individual"
                    " test timeout of {0} seconds but a timeout of {1} seconds was"
                    " requested on the command line. Forcing timeout to be {1}"
                    " seconds"
                ).format(lit_config.maxIndividualTestTime, opts.maxIndividualTestTime)
            )
            lit_config.maxIndividualTestTime = opts.maxIndividualTestTime

    determine_order(discovered_tests, opts.order)

    selected_tests = [
        t
        for t in discovered_tests
        if opts.filter.search(t.getFullName())
        and not opts.filter_out.search(t.getFullName())
    ]

    if not selected_tests:
        sys.stderr.write(
            "error: filter did not match any tests "
            "(of %d discovered).  " % len(discovered_tests)
        )
        if opts.allow_empty_runs:
            sys.stderr.write(
                "Suppressing error because '--allow-empty-runs' " "was specified.\n"
            )
            sys.exit(0)
        else:
            sys.stderr.write("Use '--allow-empty-runs' to suppress this " "error.\n")
            sys.exit(2)

    # When running multiple shards, don't include skipped tests in the xunit
    # output since merging the files will result in duplicates.
    if opts.shard:
        (run, shards) = opts.shard
        selected_tests = filter_by_shard(selected_tests, run, shards, lit_config)
        if not selected_tests:
            sys.stderr.write(
                "warning: shard does not contain any tests.  "
                "Consider decreasing the number of shards.\n"
            )
            sys.exit(0)

    selected_tests = selected_tests[: opts.max_tests]

    mark_xfail(discovered_tests, opts)

    mark_excluded(discovered_tests, selected_tests)

    start = time.time()
    run_tests(selected_tests, lit_config, opts, len(discovered_tests))
    elapsed = time.time() - start

    record_test_times(selected_tests, lit_config)

    selected_tests, discovered_tests = GoogleTest.post_process_shard_results(
        selected_tests, discovered_tests
    )

    if opts.time_tests:
        print_histogram(discovered_tests)

    print_results(discovered_tests, elapsed, opts)

    tests_for_report = selected_tests if opts.shard else discovered_tests
    for report in opts.reports:
        report.write_results(tests_for_report, elapsed)

    if lit_config.numErrors:
        sys.stderr.write("\n%d error(s) in tests\n" % lit_config.numErrors)
        sys.exit(2)

    if lit_config.numWarnings:
        sys.stderr.write("\n%d warning(s) in tests\n" % lit_config.numWarnings)

    has_failure = any(t.isFailure() for t in discovered_tests)
    if has_failure:
        if opts.ignoreFail:
            sys.stderr.write(
                "\nExiting with status 0 instead of 1 because "
                "'--ignore-fail' was specified.\n"
            )
        else:
            sys.exit(1)


def create_params(builtin_params, user_params):
    def parse(p):
        return p.split("=", 1) if "=" in p else (p, "")

    params = dict(builtin_params)
    params.update([parse(p) for p in user_params])
    return params


def print_discovered(tests, show_suites, show_tests):
    tests.sort(key=lit.reports.by_suite_and_test_path)

    if show_suites:
        tests_by_suite = itertools.groupby(tests, lambda t: t.suite)
        print("-- Test Suites --")
        for suite, test_iter in tests_by_suite:
            test_count = sum(1 for _ in test_iter)
            print("  %s - %d tests" % (suite.name, test_count))
            print("    Source Root: %s" % suite.source_root)
            print("    Exec Root  : %s" % suite.exec_root)
            features = " ".join(sorted(suite.config.available_features))
            print("    Available Features: %s" % features)
            substitutions = sorted(suite.config.substitutions)
            substitutions = ("%s => %s" % (x, y) for (x, y) in substitutions)
            substitutions = "\n".ljust(30).join(substitutions)
            print("    Available Substitutions: %s" % substitutions)

    if show_tests:
        print("-- Available Tests --")
        for t in tests:
            print("  %s" % t.getFullName())


def determine_order(tests, order):
    from lit.cl_arguments import TestOrder

    enum_order = TestOrder(order)
    if enum_order == TestOrder.RANDOM:
        import random

        random.shuffle(tests)
    elif enum_order == TestOrder.LEXICAL:
        tests.sort(key=lambda t: t.getFullName())
    else:
        assert enum_order == TestOrder.SMART, "Unknown TestOrder value"
        tests.sort(
            key=lambda t: (not t.previous_failure, -t.previous_elapsed, t.getFullName())
        )


def filter_by_shard(tests, run, shards, lit_config):
    test_ixs = range(run - 1, len(tests), shards)
    selected_tests = [tests[i] for i in test_ixs]

    # For clarity, generate a preview of the first few test indices in the shard
    # to accompany the arithmetic expression.
    preview_len = 3
    preview = ", ".join([str(i + 1) for i in test_ixs[:preview_len]])
    if len(test_ixs) > preview_len:
        preview += ", ..."
    msg = (
        f"Selecting shard {run}/{shards} = "
        f"size {len(selected_tests)}/{len(tests)} = "
        f"tests #({shards}*k)+{run} = [{preview}]"
    )
    lit_config.note(msg)
    return selected_tests


def mark_xfail(selected_tests, opts):
    for t in selected_tests:
        test_file = os.sep.join(t.path_in_suite)
        test_full_name = t.getFullName()
        if test_file in opts.xfail or test_full_name in opts.xfail:
            t.xfails += "*"
        if test_file in opts.xfail_not or test_full_name in opts.xfail_not:
            t.xfail_not = True


def mark_excluded(discovered_tests, selected_tests):
    excluded_tests = set(discovered_tests) - set(selected_tests)
    result = lit.Test.Result(lit.Test.EXCLUDED)
    for t in excluded_tests:
        t.setResult(result)


def run_tests(tests, lit_config, opts, discovered_tests):
    workers = min(len(tests), opts.workers)
    display = lit.display.create_display(opts, tests, discovered_tests, workers)

    run = lit.run.Run(
        tests, lit_config, workers, display.update, opts.max_failures, opts.timeout
    )

    display.print_header()

    interrupted = False
    error = None
    try:
        execute_in_tmp_dir(run, lit_config)
    except KeyboardInterrupt:
        interrupted = True
        error = "  interrupted by user"
    except lit.run.MaxFailuresError:
        error = "warning: reached maximum number of test failures"
    except lit.run.TimeoutError:
        error = "warning: reached timeout"

    display.clear(interrupted)
    if error:
        sys.stderr.write("%s, skipping remaining tests\n" % error)


def execute_in_tmp_dir(run, lit_config):
    # Create a temp directory inside the normal temp directory so that we can
    # try to avoid temporary test file leaks. The user can avoid this behavior
    # by setting LIT_PRESERVES_TMP in the environment, so they can easily use
    # their own temp directory to monitor temporary file leaks or handle them at
    # the buildbot level.
    tmp_dir = None
    if "LIT_PRESERVES_TMP" not in os.environ:
        import tempfile

        # z/OS linker does not support '_' in paths, so use '-'.
        tmp_dir = tempfile.mkdtemp(prefix="lit-tmp-")
        tmp_dir_envs = {k: tmp_dir for k in ["TMP", "TMPDIR", "TEMP", "TEMPDIR"]}
        os.environ.update(tmp_dir_envs)
        for cfg in {t.config for t in run.tests}:
            cfg.environment.update(tmp_dir_envs)
    try:
        run.execute()
    finally:
        if tmp_dir:
            try:
                import shutil

                shutil.rmtree(tmp_dir)
            except Exception as e:
                lit_config.warning(
                    "Failed to delete temp directory '%s', try upgrading your version of Python to fix this"
                    % tmp_dir
                )


def print_histogram(tests):
    test_times = [
        (t.getFullName(), t.result.elapsed) for t in tests if t.result.elapsed
    ]
    if test_times:
        lit.util.printHistogram(test_times, title="Tests")


def print_results(tests, elapsed, opts):
    tests_by_code = {code: [] for code in lit.Test.ResultCode.all_codes()}
    for test in tests:
        tests_by_code[test.result.code].append(test)

    for code in lit.Test.ResultCode.all_codes():
        print_group(
            sorted(tests_by_code[code], key=lambda t: t.getFullName()),
            code,
            opts.shown_codes,
        )

    print_summary(tests_by_code, opts.quiet, elapsed)


def print_group(tests, code, shown_codes):
    if not tests:
        return
    if not code.isFailure and code not in shown_codes:
        return
    print("*" * 20)
    print("{} Tests ({}):".format(code.label, len(tests)))
    for test in tests:
        print("  %s" % test.getFullName())
    sys.stdout.write("\n")


def print_summary(tests_by_code, quiet, elapsed):
    if not quiet:
        print("\nTesting Time: %.2fs" % elapsed)

    codes = [c for c in lit.Test.ResultCode.all_codes() if not quiet or c.isFailure]
    groups = [(c.label, len(tests_by_code[c])) for c in codes]
    groups = [(label, count) for label, count in groups if count]
    if not groups:
        return

    max_label_len = max(len(label) for label, _ in groups)
    max_count_len = max(len(str(count)) for _, count in groups)

    for (label, count) in groups:
        label = label.ljust(max_label_len)
        count = str(count).rjust(max_count_len)
        print("  %s: %s" % (label, count))
